@@ -3,20 +3,26 @@ import importlib
 import json
 import logging
 import os
-import sys
 import warnings
 from copy import deepcopy
 from inspect import isclass
 from typing import Any, Coroutine, Dict, List, Optional, Tuple, Type, Union, cast
 
-from tortoise import fields
+from pypika import Table
+
 from tortoise.backends.base.client import BaseDBAsyncClient
 from tortoise.backends.base.config_generator import expand_db_url, generate_config
-from tortoise.exceptions import ConfigurationError  # noqa
-from tortoise.fields import ManyToManyRelationManager  # noqa
+from tortoise.exceptions import ConfigurationError
+from tortoise.fields.relational import (
+    BackwardFKRelation,
+    BackwardOneToOneRelation,
+    ForeignKeyFieldInstance,
+    ManyToManyFieldInstance,
+    OneToOneFieldInstance,
+)
 from tortoise.filters import get_m2m_filters
 from tortoise.models import Model
-from tortoise.queryset import QuerySet  # noqa
+from tortoise.queryset import QuerySet
 from tortoise.transactions import current_transaction_map
 from tortoise.utils import generate_schema_for_client
 
@@ -27,14 +33,11 @@ except ImportError:  # pragma: nocoverage
 
 logger = logging.getLogger("tortoise")
 
-if sys.version_info < (3, 6):  # pragma: nocoverage
-    warnings.warn("Tortoise-ORM is soon going to require Python 3.6", DeprecationWarning)
-
 
 class Tortoise:
-    apps = {}  # type: Dict[str, Dict[str, Type[Model]]]
-    _connections = {}  # type: Dict[str, BaseDBAsyncClient]
-    _inited = False  # type: bool
+    apps: Dict[str, Dict[str, Type[Model]]] = {}
+    _connections: Dict[str, BaseDBAsyncClient] = {}
+    _inited: bool = False
 
     @classmethod
     def get_connection(cls, connection_name: str) -> BaseDBAsyncClient:
@@ -71,12 +74,14 @@ class Tortoise:
                     "table":                str     # DB table name
                     "abstract":             bool    # Is the model Abstract?
                     "description":          str     # Description of table (nullable)
-                    "unique_together":      [...]   # Lost of List containing field names that
+                    "unique_together":      [...]   # List of List containing field names that
                                                     #  are unique together
                     "pk_field":             {...}   # Primary key field
                     "data_fields":          [...]   # Data fields
                     "fk_fields":            [...]   # Foreign Key fields FROM this model
                     "backward_fk_fields":   [...]   # Foreign Key fields TO this model
+                    "o2o_fields":           [...]   # OneToOne fields FROM this model
+                    "backward_o2o_fields":  [...]   # OneToOne fields TO this model
                     "m2m_fields":           [...]   # Many-to-Many fields
                 }
 
@@ -92,6 +97,7 @@ class Tortoise:
                                             #  Optional: Only for pk/data fields
                     "raw_field":    str     # Name of raw field of the Foreign Key
                                             #  Optional: Only for Foreign Keys
+                    "db_field_types": dict  # DB Field types for default and DB overrides
                     "python_type":  str     # Python type
                     "generated":    bool    # Is the field generated by the DB?
                     "nullable":     bool    # Is the column nullable?
@@ -113,10 +119,10 @@ class Tortoise:
                 }
         """
 
-        def _type_name(typ) -> str:
+        def _type_name(typ: Type) -> str:
             if typ.__module__ == "builtins":
                 return typ.__name__
-            return "{}.{}".format(typ.__module__, typ.__name__)
+            return f"{typ.__module__}.{typ.__name__}"
 
         def model_name(typ: Type[Model]) -> str:
             name = typ._meta.table
@@ -124,7 +130,7 @@ class Tortoise:
                 for _name, _model in app.items():  # pragma: nobranch
                     if typ == _model:
                         name = _name
-            return "{}.{}".format(typ._meta.app, name)
+            return f"{typ._meta.app}.{name}"
 
         def type_name(typ: Any) -> Union[str, List[str]]:
             try:
@@ -141,19 +147,20 @@ class Tortoise:
             if isinstance(default, (int, float, str, bool, type(None))):
                 return default
             if callable(default):
-                return "<function {}.{}>".format(default.__module__, default.__name__)
+                return f"<function {default.__module__}.{default.__name__}>"
             return str(default)
 
         def describe_field(name: str) -> dict:
             # TODO: db_type
             field = model._meta.fields_map[name]
+            field_type = getattr(field, "model_class", field.field_type)
             desc = {
                 "name": name,
                 "field_type": field.__class__.__name__ if serializable else field.__class__,
                 "db_column": field.source_field or name,
                 "raw_field": None,
-                # "db_type": db_type,
-                "python_type": type_name(field.type) if serializable else field.type,
+                "db_field_types": field.get_db_field_types(),
+                "python_type": type_name(field_type) if serializable else field_type,
                 "generated": field.generated,
                 "nullable": field.null,
                 "unique": field.unique,
@@ -162,15 +169,21 @@ class Tortoise:
                 "description": field.description,
             }
 
+            # Delete db fields for non-db fields
+            if not desc["db_field_types"]:
+                del desc["db_field_types"]
+
             # Foreign Keys have
-            if isinstance(field, fields.ForeignKeyField):
+            if isinstance(field, (ForeignKeyFieldInstance, OneToOneFieldInstance)):
                 del desc["db_column"]
                 desc["raw_field"] = field.source_field
             else:
                 del desc["raw_field"]
 
             # These fields are entierly "virtual", so no direct DB representation
-            if isinstance(field, (fields.ManyToManyField, fields.BackwardFKRelation)):
+            if isinstance(
+                field, (ManyToManyFieldInstance, BackwardFKRelation, BackwardOneToOneRelation,),
+            ):
                 del desc["db_column"]
 
             return desc
@@ -199,6 +212,16 @@ class Tortoise:
                 for name in model._meta.fields_map.keys()
                 if name in model._meta.backward_fk_fields
             ],
+            "o2o_fields": [
+                describe_field(name)
+                for name in model._meta.fields_map.keys()
+                if name in model._meta.o2o_fields
+            ],
+            "backward_o2o_fields": [
+                describe_field(name)
+                for name in model._meta.fields_map.keys()
+                if name in model._meta.backward_o2o_fields
+            ],
             "m2m_fields": [
                 describe_field(name)
                 for name in model._meta.fields_map.keys()
@@ -218,11 +241,11 @@ class Tortoise:
 
         :param serializable:
             ``False`` if you want raw python objects,
-            ``True`` for JSON-serialisable data. (Defaults to ``True``)
+            ``True`` for JSON-serializable data. (Defaults to ``True``)
 
         :return:
             A dictionary containing the model qualifier as key,
-            and the same output as ``describe_model(…)`` as value:
+            and the same output as ``describe_model(...)`` as value:
 
             .. code-block:: python3
 
@@ -239,13 +262,13 @@ class Tortoise:
                     models.append(model)
 
         return {
-            "{}.{}".format(model._meta.app, model.__name__): cls.describe_model(model, serializable)
+            f"{model._meta.app}.{model.__name__}": cls.describe_model(model, serializable)
             for model in models
         }
 
     @classmethod
     def _init_relations(cls) -> None:
-        def get_related_model(related_app_name: str, related_model_name: str):
+        def get_related_model(related_app_name: str, related_model_name: str) -> Type[Model]:
             """
             Test, if app and model really exist. Throws a ConfigurationError with a hopefully
             helpful message. If successfull, returns the requested model.
@@ -254,13 +277,10 @@ class Tortoise:
                 return cls.apps[related_app_name][related_model_name]
             except KeyError:
                 if related_app_name not in cls.apps:
-                    raise ConfigurationError(
-                        "No app with name '{}' registered.".format(related_app_name)
-                    )
+                    raise ConfigurationError(f"No app with name '{related_app_name}' registered.")
                 raise ConfigurationError(
-                    "No model with name '{}' registered in app '{}'.".format(
-                        related_model_name, related_app_name
-                    )
+                    f"No model with name '{related_model_name}' registered in"
+                    f" app '{related_app_name}'."
                 )
 
         def split_reference(reference: str) -> Tuple[str, str]:
@@ -285,20 +305,42 @@ class Tortoise:
             for model_name, model in app.items():
                 if model._meta._inited:
                     continue
-                else:
-                    model._meta._inited = True
+                model._meta._inited = True
                 if not model._meta.table:
                     model._meta.table = model.__name__.lower()
 
+                pk_attr_changed = False
+
+                # TODO: refactor to share logic between FK & O2O
                 for field in model._meta.fk_fields:
-                    fk_object = cast(fields.ForeignKeyField, model._meta.fields_map[field])
+                    fk_object = cast(ForeignKeyFieldInstance, model._meta.fields_map[field])
                     reference = fk_object.model_name
                     related_app_name, related_model_name = split_reference(reference)
                     related_model = get_related_model(related_app_name, related_model_name)
 
-                    key_field = "{}_id".format(field)
-                    key_fk_object = deepcopy(related_model._meta.pk)
+                    if fk_object.to_field:
+                        related_field = related_model._meta.fields_map.get(fk_object.to_field, None)
+                        if related_field:
+                            if related_field.unique:
+                                key_fk_object = deepcopy(related_field)
+                                fk_object.to_field_instance = related_field
+                            else:
+                                raise ConfigurationError(
+                                    f'field "{fk_object.to_field}" in model'
+                                    f' "{related_model_name}" is not unique'
+                                )
+                        else:
+                            raise ConfigurationError(
+                                f'there is no field named "{fk_object.to_field}"'
+                                f' in model "{related_model_name}"'
+                            )
+                    else:
+                        key_fk_object = deepcopy(related_model._meta.pk)
+                        fk_object.to_field_instance = related_model._meta.pk
+
+                    key_field = f"{field}_id"
                     key_fk_object.pk = False
+                    key_fk_object.unique = False
                     key_fk_object.index = fk_object.index
                     key_fk_object.default = fk_object.default
                     key_fk_object.null = fk_object.null
@@ -313,49 +355,116 @@ class Tortoise:
                         key_fk_object.source_field = key_field
                     model._meta.add_field(key_field, key_fk_object)
 
-                    fk_object.type = related_model
+                    fk_object.model_class = related_model
                     backward_relation_name = fk_object.related_name
-                    if not backward_relation_name:
-                        backward_relation_name = "{}s".format(model._meta.table)
-                    if backward_relation_name in related_model._meta.fields:
-                        raise ConfigurationError(
-                            'backward relation "{}" duplicates in model {}'.format(
-                                backward_relation_name, related_model_name
+                    if backward_relation_name is not False:
+                        if not backward_relation_name:
+                            backward_relation_name = f"{model._meta.table}s"
+                        if backward_relation_name in related_model._meta.fields:
+                            raise ConfigurationError(
+                                f'backward relation "{backward_relation_name}" duplicates in'
+                                f" model {related_model_name}"
                             )
+                        fk_relation = BackwardFKRelation(
+                            model, f"{field}_id", fk_object.null, fk_object.description,
                         )
-                    fk_relation = fields.BackwardFKRelation(
-                        model, "{}_id".format(field), fk_object.description
-                    )
-                    related_model._meta.add_field(backward_relation_name, fk_relation)
+                        fk_relation.to_field_instance = fk_object.to_field_instance
+                        related_model._meta.add_field(
+                            cast(str, backward_relation_name), fk_relation
+                        )
+
+                for field in model._meta.o2o_fields:
+                    o2o_object = cast(OneToOneFieldInstance, model._meta.fields_map[field])
+                    reference = o2o_object.model_name
+                    related_app_name, related_model_name = split_reference(reference)
+                    related_model = get_related_model(related_app_name, related_model_name)
+
+                    if o2o_object.to_field:
+                        related_field = related_model._meta.fields_map.get(
+                            o2o_object.to_field, None
+                        )
+                        if related_field:
+                            if related_field.unique:
+                                key_o2o_object = deepcopy(related_field)
+                                o2o_object.to_field_instance = related_field
+                            else:
+                                raise ConfigurationError(
+                                    f'field "{o2o_object.to_field}" in model'
+                                    f' "{related_model_name}" is not unique'
+                                )
+                        else:
+                            raise ConfigurationError(
+                                f'there is no field named "{o2o_object.to_field}"'
+                                f' in model "{related_model_name}"'
+                            )
+                    else:
+                        key_o2o_object = deepcopy(related_model._meta.pk)
+                        o2o_object.to_field_instance = related_model._meta.pk
+
+                    key_field = f"{field}_id"
+                    key_o2o_object.pk = o2o_object.pk
+                    key_o2o_object.index = o2o_object.index
+                    key_o2o_object.default = o2o_object.default
+                    key_o2o_object.null = o2o_object.null
+                    key_o2o_object.unique = o2o_object.unique
+                    key_o2o_object.generated = o2o_object.generated
+                    key_o2o_object.reference = o2o_object
+                    key_o2o_object.description = o2o_object.description
+                    if o2o_object.source_field:
+                        key_o2o_object.source_field = o2o_object.source_field
+                        o2o_object.source_field = key_field
+                    else:
+                        o2o_object.source_field = key_field
+                        key_o2o_object.source_field = key_field
+                    model._meta.add_field(key_field, key_o2o_object)
+
+                    o2o_object.model_class = related_model
+                    backward_relation_name = o2o_object.related_name
+                    if backward_relation_name is not False:
+                        if not backward_relation_name:
+                            backward_relation_name = f"{model._meta.table}"
+                        if backward_relation_name in related_model._meta.fields:
+                            raise ConfigurationError(
+                                f'backward relation "{backward_relation_name}" duplicates in'
+                                f" model {related_model_name}"
+                            )
+                        o2o_relation = BackwardOneToOneRelation(
+                            model, f"{field}_id", null=True, description=o2o_object.description,
+                        )
+                        o2o_relation.to_field_instance = o2o_object.to_field_instance
+                        related_model._meta.add_field(
+                            cast(str, backward_relation_name), o2o_relation
+                        )
+
+                    if o2o_object.pk:
+                        pk_attr_changed = True
+                        model._meta.pk_attr = key_field
 
                 for field in list(model._meta.m2m_fields):
-                    m2m_object = cast(fields.ManyToManyField, model._meta.fields_map[field])
+                    m2m_object = cast(ManyToManyFieldInstance, model._meta.fields_map[field])
                     if m2m_object._generated:
                         continue
 
                     backward_key = m2m_object.backward_key
                     if not backward_key:
-                        backward_key = "{}_id".format(model._meta.table)
+                        backward_key = f"{model._meta.table}_id"
                         if backward_key == m2m_object.forward_key:
-                            backward_key = "{}_rel_id".format(model._meta.table)
+                            backward_key = f"{model._meta.table}_rel_id"
                         m2m_object.backward_key = backward_key
 
                     reference = m2m_object.model_name
                     related_app_name, related_model_name = split_reference(reference)
                     related_model = get_related_model(related_app_name, related_model_name)
 
-                    m2m_object.type = related_model
+                    m2m_object.model_class = related_model
 
                     backward_relation_name = m2m_object.related_name
                     if not backward_relation_name:
-                        backward_relation_name = m2m_object.related_name = "{}_through".format(
-                            model._meta.table
-                        )
+                        backward_relation_name = m2m_object.related_name = f"{model._meta.table}s"
                     if backward_relation_name in related_model._meta.fields:
                         raise ConfigurationError(
-                            'backward relation "{}" duplicates in model {}'.format(
-                                backward_relation_name, related_model_name
-                            )
+                            f'backward relation "{backward_relation_name}" duplicates in'
+                            f" model {related_model_name}"
                         )
 
                     if not m2m_object.through:
@@ -365,23 +474,24 @@ class Tortoise:
                             else related_model.__name__.lower()
                         )
 
-                        m2m_object.through = "{}_{}".format(
-                            model._meta.table, related_model_table_name
-                        )
+                        m2m_object.through = f"{model._meta.table}_{related_model_table_name}"
 
-                    m2m_relation = fields.ManyToManyField(
-                        "{}.{}".format(app_name, model_name),
+                    m2m_relation = ManyToManyFieldInstance(
+                        f"{app_name}.{model_name}",
                         m2m_object.through,
                         forward_key=m2m_object.backward_key,
                         backward_key=m2m_object.forward_key,
                         related_name=field,
-                        type=model,
+                        field_type=model,
                         description=m2m_object.description,
                     )
                     m2m_relation._generated = True
                     model._meta.filters.update(get_m2m_filters(field, m2m_object))
                     related_model._meta.add_field(backward_relation_name, m2m_relation)
 
+                if pk_attr_changed:
+                    model._meta.finalise_pk()
+
     @classmethod
     def _discover_client_class(cls, engine: str) -> BaseDBAsyncClient:
         # Let exception bubble up for transparency
@@ -390,9 +500,7 @@ class Tortoise:
         try:
             client_class = engine_module.client_class  # type: ignore
         except AttributeError:
-            raise ConfigurationError(
-                'Backend for engine "{}" does not implement db client'.format(engine)
-            )
+            raise ConfigurationError(f'Backend for engine "{engine}" does not implement db client')
         return client_class
 
     @classmethod
@@ -400,7 +508,7 @@ class Tortoise:
         try:
             module = importlib.import_module(models_path)
         except ImportError:
-            raise ConfigurationError('Module "{}" not found'.format(models_path))
+            raise ConfigurationError(f'Module "{models_path}" not found')
         discovered_models = []
         possible_models = getattr(module, "__models__", None)
         try:
@@ -416,6 +524,8 @@ class Tortoise:
                 attr._meta.app = app_label
                 attr._meta.finalise_pk()
                 discovered_models.append(attr)
+        if not discovered_models:
+            warnings.warn(f'Module "{models_path}" has no models', RuntimeWarning, stacklevel=4)
         return discovered_models
 
     @classmethod
@@ -424,7 +534,7 @@ class Tortoise:
             if isinstance(info, str):
                 info = expand_db_url(info)
             client_class = cls._discover_client_class(info.get("engine"))
-            db_params = deepcopy(info["credentials"])
+            db_params = info["credentials"].copy()
             db_params.update({"connection_name": name})
             connection = client_class(**db_params)  # type: ignore
             if create_db:
@@ -444,7 +554,7 @@ class Tortoise:
                         info.get("default_connection", "default"), name
                     )
                 )
-            app_models = []  # type: List[Type[Model]]
+            app_models: List[Type[Model]] = []
             for module in info["models"]:
                 app_models += cls._discover_models(module, name)
 
@@ -463,7 +573,7 @@ class Tortoise:
     def _get_config_from_config_file(cls, config_file: str) -> dict:
         _, extension = os.path.splitext(config_file)
         if extension in (".yml", ".yaml"):
-            import yaml
+            import yaml  # pylint: disable=C0415
 
             with open(config_file, "r") as f:
                 config = yaml.safe_load(f)
@@ -472,7 +582,7 @@ class Tortoise:
                 config = json.load(f)
         else:
             raise ConfigurationError(
-                "Unknown config extension {}, only .yml and .json are supported".format(extension)
+                f"Unknown config extension {extension}, only .yml and .json are supported"
             )
         return config
 
@@ -481,6 +591,7 @@ class Tortoise:
         for app in cls.apps.values():
             for model in app.values():
                 model._meta.finalise_model()
+                model._meta.basetable = Table(model._meta.table)
                 model._meta.basequery = model._meta.db.query_class.from_(model._meta.table)
                 model._meta.basequery_all_fields = model._meta.basequery.select(
                     *model._meta.db_fields
@@ -609,7 +720,7 @@ class Tortoise:
         for app in cls.apps.values():
             for model in app.values():
                 model._meta.default_connection = None
-        cls.apps = {}
+        cls.apps.clear()
         current_transaction_map.clear()
 
     @classmethod
@@ -670,4 +781,4 @@ def run_async(coro: Coroutine) -> None:
         loop.run_until_complete(Tortoise.close_connections())
 
 
-__version__ = "0.13.1"
+__version__ = "0.15.20"
